///////////////////////////////////////////////////////////////////////////////
//CS 441 - Distributed Object Programming - Fall 2004                        //
//MP 1                                                                       //
//                                                                           //
//Name : Jin Soon Lim                                                        //
//CS id: jlim                                                                //
//Due date: 09/21/2004                                                       //
///////////////////////////////////////////////////////////////////////////////

//source: Dbs.java                                                           //
//tested at bert.cs.uic.edu                                                  //

import java.io.*;
import java.net.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.table.*;

public class Dbs extends JFrame implements ActionListener {
    final int PORT = 15703;
    final int NUM_RECORDS = 6;  // total # of records
    
    static Dbs dbs;
    ThreadGroup dbsThreadGrp;
    Map clientlist;
    ServerSocket server;
    DbTableModel dbtm;
    private Records[] records;

    JTextArea textbox;
    JMenuItem m_clrl, m_stop, m_exit, m_about;
    
	private Dbs() {
        // initialize records
        records = new Records[NUM_RECORDS];
        for (int i=0; i<NUM_RECORDS; i++)
            records[i] = new Records(i, i*3 + "", 0);
    
        // create user interface
        dbtm = new DbTableModel();
        gui();
    }

    private void start() {
        try {
            int clientID = 0;
            server = new ServerSocket(PORT);                   // create a socket for client connection

            setTitle("Dbs - Running");
            println("ServerSocket is bound to port " + PORT + ".");
            
            clientlist = new HashMap();                        // map for client list
            dbsThreadGrp = new ThreadGroup("Dbsthread group"); // thread group of client threads 
            
            while (true) {
                // create a thread for new client
                DbsThread dt = new DbsThread(server.accept(), clientID + "");
                Thread t = new Thread(dbsThreadGrp, dt, clientID + "");
            	t.start();
                
                clientlist.put(clientID + "", dt);             // add the thread to client list with clientID
                println("Client " + clientID + " is connected.");
                clientID++;
            }
        }
        catch (IOException e) {
            System.err.println(e.getMessage());
            if (server != null && !server.isClosed())
                stop();                                        // stop server
        }
    }
    
    private void stop() {
        try {
    		// make all clients stop
            sendPacketToAll(new Dbpacket(Dbpacket.OP_SERVER_SHUTDOWN));
            
            // close server socket
            server.close();
            println("ServerSocket is closed.");
            setTitle("Dbs - Stopped");
            m_stop.setEnabled(false);
            m_clrl.setEnabled(false);
        }
        catch (IOException e) {
            System.err.println(e.getMessage());
        }
    }
    
    private void exit() {
    	// stop and exit program
        stop();
        System.exit(0);
    }
    
    private void notifyToAll(String str) {
        // send a message to all clients
        sendPacketToAll(new Dbpacket(Dbpacket.OP_SEND_MSG, str));
    }
    
    private void sendPacketToAll(Dbpacket p) {
        int nt = dbsThreadGrp.activeCount();    // get number of active threads
        Thread[] threadlist = new Thread[nt];   // create a temporary array
        dbsThreadGrp.enumerate(threadlist);     // copy all active threads to the array
        
        // send a packet to all clients
        for (int i = 0; i < nt; i++) {
            Object obj = clientlist.get(threadlist[i].getName()); // find a client with the clientID
            ((DbsThread)obj).sendPacket(p);                       // send a packet to the client
        }
    }

    // function for the clear all locks command of the server
    private void clearAllLocks() {
    	synchronized(records) {
    		// clear all locks on the master record
    		for (int i = 0; i < records.length; i++)
            	records[i].lock = 0;
  
    		// get active clients
    		int nt = dbsThreadGrp.activeCount();
    		Thread[] threadlist = new Thread[nt];
    		dbsThreadGrp.enumerate(threadlist);
    	
    		// clear all locks in active clients
    		for (int i = 0; i < nt; i++) {
    			Object obj = clientlist.get(threadlist[i].getName());
    			((DbsThread)obj).clearAllLocksInClient();
    		}
        }
        
        dbtm.fireTableDataChanged();  // update the record table in gui
        println("All locks are cleared.");
    }
    
    private void println(String str) {
        // print a string to the server text area
        textbox.append(str + "\n");
        textbox.setCaretPosition(textbox.getDocument().getLength()); 
    }
    
    private void gui() {
        setSize(240, 320);
        JPanel mainpanel = new JPanel(new GridLayout(2, 1));

        // create a text area
        textbox = new JTextArea();
        textbox.setEditable(false);
        JScrollPane scrolltextbox = new JScrollPane(textbox);
            
        // create a table 
        JTable rtable = new JTable(dbtm);
        JScrollPane scrolltable = new JScrollPane(rtable);
        
        // create menu
        JMenuBar menubar = new JMenuBar();
        JMenu menu = new JMenu("Server");
        m_clrl = new JMenuItem("Clear All Locks");
        m_clrl.addActionListener(this);
        m_stop = new JMenuItem("Stop");
        m_stop.addActionListener(this);
        m_exit = new JMenuItem("Exit");
        m_exit.addActionListener(this);
        menu.add(m_clrl);
        menu.add(m_stop);
        menu.add(m_exit);
        menubar.add(menu);
        menu = new JMenu("Help");
        m_about = new JMenuItem("About");
        m_about.addActionListener(this);
        menu.add(m_about);
        menubar.add(menu);

        mainpanel.add(scrolltable);
        mainpanel.add(scrolltextbox);
        setJMenuBar(menubar);
        setContentPane(mainpanel);
        setVisible(true);
    }
    
    public void actionPerformed (ActionEvent e) {
        // handle action events
        Object obj = e.getSource();

        if (obj == m_clrl) {
        	clearAllLocks();
        }
        else if (obj == m_stop) {
            stop();
        }
        else if (obj == m_exit) {
            exit();
        }
        else if (obj == m_about) {
            JOptionPane.showMessageDialog(this, "CS 441 MP 1\nJin Soon Lim\n09/21/2004");
        }
    }
    
    protected void processWindowEvent(WindowEvent e) {
        // for processing program exit
        if (e.getID() == WindowEvent.WINDOW_CLOSING)
            exit();
        else
            super.processWindowEvent(e);
    }

    // class for client threads
    private class DbsThread implements Runnable {
        String id;                       // client id
        boolean active = false;          // client state
        boolean readlock[], writelock[]; // locks in a client
        
        Socket socket = null;
        ObjectInputStream in = null;
        ObjectOutputStream out = null;
        
        public DbsThread(Socket socket, String id) {
            this.socket = socket;
            this.id = id;
            
            // initialize locks
            int nr = dbs.records.length;
            readlock = new boolean[nr];
            writelock = new boolean[nr];
            for (int i = 0; i < nr; i++)
                readlock[i] = writelock[i] = false;
        }

        public void run() {
                active = true;
            try {
                // set input/output stream
                in = new ObjectInputStream(socket.getInputStream());
                out = new ObjectOutputStream(socket.getOutputStream());
                
                // send id to the client
                sendPacket(new Dbpacket(Dbpacket.OP_CLIENT_INIT, id));
                
                // wait for packets from the client
                while (active) {
                	unpacker((Dbpacket)in.readObject());  // unpack the packet recieved
                }
            }
            catch (IOException e) {
                System.err.println(e.getMessage());
            }
            catch (ClassNotFoundException e) {
                System.err.println(e.getMessage());
            }
            finally {
                try {
                    // close io streams and socket
                    if (in != null) in.close();
                    if (out != null) out.close();
                    if (socket != null && !socket.isClosed()) socket.close();
                }
                catch (IOException e) {
                    System.err.println(e.getMessage());
                }
            }
        }
        
        // unpack the packet recieved
        private void unpacker(Dbpacket p) {
            // verify valid rid for related opcodes
            if (p.getOpcode() <= Dbpacket.OP_CLEAR_LOCK) {
                if (p.getRid() >= dbs.records.length || p.getRid() < 0) {
                    sendMsg("Wrong record number.");
                    return;
                }
            }
            if (p.getOpcode() == Dbpacket.OP_WRITE) {
            	if (p.getValue() == null || p.getValue().length() == 0) {
            		sendMsg("Wrong input value.");
                    return;
                }
            }
            
            // process opcodes
            switch (p.getOpcode()) {
                // read a record
                case Dbpacket.OP_READ:
                    read(p.getRid());
                    break;
                
                // write a record
                case Dbpacket.OP_WRITE:
                    write(p.getRid(), p.getValue());
                    break;
                
                // set a read lock
                case Dbpacket.OP_SET_RL:
                	if (setrl(p.getRid()))
                        sendMsg("A read lock is set on record " + p.getRid() + ".");
                    else
                        sendMsg("Failed. There is a write lock on record " + p.getRid() + ".");
                    break;
                
                // set a write lock
                case Dbpacket.OP_SET_WL:
                	if (setwl(p.getRid()))
                		sendMsg("A write lock is set on record " + p.getRid() + ".");
                	else
                		sendMsg("Failed. There is a write lock on record " + p.getRid() + ".");
                    break;
                
                // clear a lock
                case Dbpacket.OP_CLEAR_LOCK:
                    if (!clearlock(p.getRid()))
                        sendMsg("You don't have any locks to clear on record " + p.getRid() + ".");
                    break;
                
                // client exit
                case Dbpacket.OP_CLIENT_EXIT:
                    // release all locks before exit
                    for (int i=0; i<readlock.length; i++)
                    	clearlock(i);
                
                    active = false;  // stop the client thread
                    dbs.println("Client " + id + " is disconnected.");
                    break;
                
                // print a message recieved 
                case Dbpacket.OP_SEND_MSG:
                    dbs.println(p.getValue());
                    break;
                
                default:
                    dbs.println("Error: undefined packet.");
            }
            dbtm.fireTableDataChanged(); // update the record table in gui
        }
    
        private void read(int rid) {
            // check and set a read lock
            setrl(rid);
            
            if (readlock[rid]) {
                // read the record
                sendMsg("Record " + rid + " is " + dbs.records[rid].value + ".");

                // release the read lock
                synchronized(dbs.records[rid]) {
                    dbs.records[rid].lock--;
                    readlock[rid] = false;
                }
            }
            else
                sendMsg("Reading is failed. There is a write lock on record " + rid + ".");
        }
        
        private void write(int rid, String value) {
            // check and set a write lock
            setwl(rid);
            
            if (writelock[rid]) {
                // write the record
                dbs.records[rid].value = value;
                sendMsg("Record " + rid + " is updated.");
                dbs.println("Record " + rid + " is updated.");

                // release the write lock
                synchronized(dbs.records[rid]) {
                    dbs.records[rid].lock = 0;
                    writelock[rid] = false;
                }
            }
            else
                sendMsg("Writing is failed. There is a lock on record " + rid + ".");
        }        
        
        private boolean setrl(int rid) {
        	synchronized(dbs.records[rid]) {
                if (!readlock[rid] && (dbs.records[rid].lock >= 0)) { // if it has not a read lock
                    dbs.records[rid].lock++;                          // and no writer exists,
                    readlock[rid] = true;                             // get a lock.
                }
            }
                return (readlock[rid])? true:false;
        }
        
        private boolean setwl(int rid) {
            synchronized(dbs.records[rid]) {
                if (!writelock[rid] && (dbs.records[rid].lock == 0)) { // if it has not a write lock
                    dbs.records[rid].lock = -1;                        // and the record is unlocked,  
                    writelock[rid] = true;                             // get a lock.
                }
            }
                return (writelock[rid])? true:false;
        }
        
        private boolean clearlock(int rid) {
            // check if it has a lock
            if (!readlock[rid] && !writelock[rid]) 
                return false;
            
            synchronized(dbs.records[rid]) {
                // clear a read lock
                if (readlock[rid]) {
                    dbs.records[rid].lock--;
                    readlock[rid] = false;
                    sendMsg("A read lock is cleared.");
                }
                
                // clear a write lock
                if (writelock[rid]) {
                    dbs.records[rid].lock = 0;
                    writelock[rid] = false;
                    sendMsg("A write lock is cleared.");
                }
            }
            return true;
        }
        
        private void clearAllLocksInClient() {
            // clear all locks the client has
            for (int i = 0; i < readlock.length; i++) 
                readlock[i] = writelock[i] = false;
            
            sendMsg("All locks are cleared by the server.");
        }
            
        private void sendMsg(String str) {
            // send a message to the client
            Dbpacket p = new Dbpacket(Dbpacket.OP_SEND_MSG, str);
            sendPacket(p);
        }
        
        private void sendPacket(Dbpacket p) {
            // send a packet to the client
            try {
                out.writeObject(p);
                out.flush();
            }
            catch (IOException e) {
                System.err.println(e.getMessage());
            }
        }
	} // end of DbsThread class

    // class for record table model in gui
    private class DbTableModel extends AbstractTableModel { 
        String header[] = new String[] {"Id", "Value", "Lock"};

        // overide methods
        public DbTableModel() { }
        public int getRowCount() { return records.length; }
        public int getColumnCount() { return header.length; }
        public String getColumnName(int c) { return header[c]; }
        public boolean isCellEditable(int r, int c) { return false; }
        public Object getValueAt(int r, int c) {
            Object obj = null;
            if (c == 0)
            	obj = new Integer(records[r].id);
            else if (c == 1)
            	obj = new String(records[r].value);
            else if (c == 2) {
                if (records[r].lock == 0)
                    obj = new String("Unlocked");
                else if (records[r].lock > 0)
                    obj = new String("Read-lock");
                else
                    obj = new String("Write-lock");
            }
            return obj;
        }
    } // end of DbTableModel class
    
    // Db record class
    private class Records {
        int id, lock;  // unlocked == 0, read-lock > 0, write-lock == -1
        String value;
        
        public Records(int id, String value, int lock) {
            this.id = id;
            this.value = value;
            this.lock = lock;
        }
    } // end of Records class
    
    public static void main(String[] args) {
        dbs = new Dbs();
        dbs.start();    // start the server
    }
} // end of Dbs class